//! Simple calendar applications.
mod app;
mod ui;
mod model;

use ui::{calendar, controls::Controls};
use app::Message;
use model::{
	events::EventsCollection,
	ical_bridge::load_calendar,
};
use chrono::{Datelike, NaiveDate, Months, Utc, Days};
use calendar::{CalendarParams, CalendarView };
use calendar::CalendarViewMode as ViewMode;
use std::path;
use clap::Parser;
use log::info;
use simplelog::{SimpleLogger, Config};
use iced::{
    Alignment,
    Command,
    Element,
    Length,
    widget::{
	    Column,
	    Container,
	},
};

#[cfg(feature = "tracing")]
extern crate lttng_ust;

#[cfg(feature = "tracing")]
use lttng_ust::import_tracepoints;

#[cfg(feature = "tracing")]
import_tracepoints!(
    concat!(env!("OUT_DIR"), "/tracepoints.rs"),
    tracepoints
);

#[derive(Parser, Debug)]
#[command(version, about)]
struct CliArgs {
    files: Vec<String>,
}

pub fn main() -> iced::Result {
	set_up_logger();

    let args = CliArgs::parse();
    let calendar_files: Vec<_> = args.files.iter().map(|arg| {
    	info!("CalFile: {}", arg);
    	path::PathBuf::from(arg)
    }).collect();

	let app_factory = move || { CalendarApp::with_files(calendar_files.clone())};
	iced::program("Calendar", CalendarApp::update, CalendarApp::view).run_with(app_factory)
}

fn set_up_logger() {
	SimpleLogger::init(log::LevelFilter::Info, Config::default()).unwrap();
}

#[derive(Default)]
struct CalendarApp {
    view_date: NaiveDate,
    controls: Controls,
    events: EventsCollection,
}

impl std::fmt::Display for ViewMode {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "{}",
            match self {
                ViewMode::Week => "Week",
                ViewMode::Month => "Month",
                ViewMode::Year => "Year",
            }
        )
    }
}

impl CalendarApp {
	fn with_files(calendar_paths: std::vec::Vec<path::PathBuf>) -> Self {
        let view_date = Utc::now().date_naive();
        let mut app = CalendarApp {
            view_date,
            events: EventsCollection::new(),
            ..CalendarApp::default()
        };
        for calendar_path in calendar_paths {
            load_calendar(&calendar_path, &mut app.events);
        }
        app
	}

    fn update(&mut self, message: Message) -> Command<Message> {
        match message {
            Message::PrevWeek => {
                self.view_date = self.view_date - Days::new(7);
            }
            Message::NextWeek => {
                self.view_date = self.view_date + Days::new(7);
            }
            Message::PrevMonth => {
                self.view_date = self.view_date - Months::new(1);
            }
            Message::NextMonth => {
                self.view_date = self.view_date + Months::new(1);
            }
            Message::PrevYear => {
                self.view_date = self.view_date - Months::new(12);
            }
            Message::NextYear => {
                self.view_date = self.view_date + Months::new(12);
            }
            Message::ViewModeSelected(mode) => {
                self.controls.mode = Some(mode);
            }
        }
        Command::none()
    }

    fn view(&self) -> Element<Message> {
        #[cfg(feature = "tracing")]
        tracepoints::calendar::view_entry();
        let content = Column::new()
            .align_items(Alignment::Center)
            .push(self.controls.view(self.view_date, self.view_date.year()))
            .push(CalendarView::new(self.controls.mode.unwrap_or(ViewMode::Year), &CalendarParams::new(), self.view_date, &self.events))
            ;

        let container = Container::new(content)
            .width(Length::Fill)
            .height(Length::Fill)
            .center_x()
            .center_y()
            .into();

        #[cfg(feature = "tracing")]
        tracepoints::calendar::view_exit();

        container
    }
}
